package com.xy.service;

import cn.hutool.json.JSONUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.github.yitter.idgen.YitIdHelper;
import com.xy.collections.list.JArrayList;
import com.xy.collections.list.JList;
import com.xy.collections.map.JHashMap;
import com.xy.collections.map.JMap;
import com.xy.device.EnumAlgorithmPayConfig;
import com.xy.device.EnumDeviceAlgorithmChargingHistoryStatus;
import com.xy.device.EnumDeviceAlgorithmChargingType;
import com.xy.dto.DeviceAlgorithmChargingDto;
import com.xy.dto.DeviceInfoDto;
import com.xy.entity.DeviceAlgorithmCharging;
import com.xy.entity.DeviceAlgorithmChargingHistory;
import com.xy.entity.SysDictRedis;
import com.xy.error.CommRuntimeException;
import com.xy.mapper.DeviceAlgorithmChargingMapper;
import com.xy.utils.*;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.AllArgsConstructor;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import static com.xy.utils.PlusBeans.toPageBean;


/**
 * <p>
 * 设备算法计费表 服务实现类
 * </p>
 *
 * @author lijin
 * @since 2023-06-29
 */
@Service
@AllArgsConstructor
@Api(tags = "设备算法计费表")
public class DeviceAlgorithmChargingServiceImpl extends ServiceImpl<DeviceAlgorithmChargingMapper, DeviceAlgorithmCharging> implements DeviceAlgorithmChargingService {

    private DeviceAlgorithmChargingHistoryServiceImpl deviceAlgorithmChargingHistoryService;

    private DeviceInfoService deviceInfoService;

    @PostMapping("page")
    @ApiOperation("分页查询")
    public R<PageBean<DeviceAlgorithmChargingDto.Vo>> page(@RequestBody @Validated DeviceAlgorithmChargingDto.Page page) {
        LambdaQueryWrapper<DeviceAlgorithmCharging> lambdaQueryWrapper = new MybatisPlusQuery().eqWrapper(page, DeviceAlgorithmCharging.class)
                .build()
                .orderByDesc(DeviceAlgorithmCharging::getTimeout, DeviceAlgorithmCharging::getUnusedSize);
        IPage<DeviceAlgorithmCharging> iPage = page(PlusBeans.toIPage(page.getPage()), lambdaQueryWrapper);
        return R.ok(toPageBean(DeviceAlgorithmChargingDto.Vo.class, iPage));
    }

    @PostMapping("check")
    @ApiOperation("检查")
    public R check(@RequestBody @Validated DeviceAlgorithmChargingDto.Check check) {
        DeviceInfoDto.Vo deviceInfo = deviceInfoService.obj(new DeviceInfoDto.Obj().setDeviceId(check.getDeviceId())).getData();
        List<DeviceAlgorithmCharging> deviceAlgorithmChargings = list(new LambdaQueryWrapper<DeviceAlgorithmCharging>()
                .eq(DeviceAlgorithmCharging::getDeviceId, deviceInfo.getDeviceId())
                .eq(DeviceAlgorithmCharging::getAlgorithmId, deviceInfo.getAlgorithmId())
                .gt(DeviceAlgorithmCharging::getUnusedSize, 0)
                .gt(DeviceAlgorithmCharging::getTimeout, 0)
                .in(DeviceAlgorithmCharging::getBeginTime, Arrays.asList(DataTime.getSring("yyyyMM"), -1))
        );
        if (!Emptys.check(deviceAlgorithmChargings)) {
            throw new CommRuntimeException("算法费欠费");
        }
        return R.ok();
    }

    @Override
    @ApiOperation("购买")
    public R pay(DeviceAlgorithmChargingDto.Pay pay) {
        JList<DeviceAlgorithmChargingDto.Pay.DeviceAlgorithmCharging> deviceAlgorithmChargings = new JArrayList<>(pay.getDeviceAlgorithmChargings());
        LocalDateTime now = LocalDateTime.now();
        List<DeviceAlgorithmChargingHistory> deviceAlgorithmChargingHistories = new ArrayList<>(deviceAlgorithmChargings.size());
        deviceAlgorithmChargings.forEach(deviceAlgorithmCharging -> {
            DeviceAlgorithmChargingHistory deviceAlgorithmChargingHistory = new DeviceAlgorithmChargingHistory()
                    .setId(YitIdHelper.nextId())
                    .setOrderId(pay.getOrderId())
                    .setDeviceId(deviceAlgorithmCharging.getDeviceId())
                    .setAlgorithmId(deviceAlgorithmCharging.getAlgorithmId())
                    .setChargingSize(deviceAlgorithmCharging.getChargingSize())
                    .setChargingMoney(deviceAlgorithmCharging.getMoney())
                    .setType(deviceAlgorithmCharging.getType())
                    .setName(deviceAlgorithmCharging.getName())
                    .setChargingType(pay.getPayType())
                    .setCreateTime(now)
                    .setUpdateTime(now);
            deviceAlgorithmChargingHistories.add(deviceAlgorithmChargingHistory);
        });
        //添加设备算法计费历史表
        if (Emptys.check(deviceAlgorithmChargingHistories)) {
            deviceAlgorithmChargingHistoryService.saveBatch(deviceAlgorithmChargingHistories);
        }
        return R.ok();
    }

    @Override
    @ApiOperation("购买回调")
    public R payNotice(DeviceAlgorithmChargingDto.PayNotice payNotice) {
        List<DeviceAlgorithmChargingHistory> list = deviceAlgorithmChargingHistoryService.list(new LambdaQueryWrapper<DeviceAlgorithmChargingHistory>().eq(DeviceAlgorithmChargingHistory::getOrderId, payNotice.getOrderId()));
        if (!Emptys.check(list)) {
            return R.ok();
        }
        JList<DeviceAlgorithmChargingHistory> deviceAlgorithmChargingHistories = new JArrayList<>(list);
        JList<DeviceAlgorithmChargingHistory> infs = new JArrayList<>();
        JList<DeviceAlgorithmChargingHistory> gives = new JArrayList<>();
        LocalDateTime now = LocalDateTime.now();
        int yyyyMMdd = Integer.valueOf(DataTime.getSring("yyyyMMdd"));
        //获取算法包月字典
        Map<String, SysDictRedis> algorithmMoonConfigMaps = SysDictUtils.get(EnumAlgorithmPayConfig.Code.CODE.getCode());
        //获取类型字典
        Map<String, SysDictRedis> deviceAlgorithmChargingTypeMaps = SysDictUtils.get(EnumDeviceAlgorithmChargingType.Code.CODE.getCode());
        //处理有效期
        deviceAlgorithmChargingHistories.forEach(deviceAlgorithmChargingHistory -> {
            Integer type = deviceAlgorithmChargingHistory.getType();
            SysDictRedis deviceAlgorithmChargingType2 = deviceAlgorithmChargingTypeMaps.get(EnumDeviceAlgorithmChargingType.N_2.getCode());
            if (type.equals(Integer.valueOf(deviceAlgorithmChargingType2.getValue()))) {
                infs.add(deviceAlgorithmChargingHistory);
                return;
            }
            SysDictRedis deviceAlgorithmChargingType3 = deviceAlgorithmChargingTypeMaps.get(EnumDeviceAlgorithmChargingType.N_3.getCode());
            if (type.equals(Integer.valueOf(deviceAlgorithmChargingType3.getValue()))) {
                gives.add(deviceAlgorithmChargingHistory);
                return;
            }
            SysDictRedis algorithmMoonConfigDict = algorithmMoonConfigMaps.get(deviceAlgorithmChargingHistory.getAlgorithmId().toString());
            DeviceAlgorithmChargingDto.AlgorithmMoonConfig algorithmMoonConfig = JSONUtil.parseObj(algorithmMoonConfigDict.getValue()).toBean(DeviceAlgorithmChargingDto.AlgorithmMoonConfig.class);
            DeviceAlgorithmChargingDto.AlgorithmMoonConfig.Moon moon = algorithmMoonConfig.getMoon();
            Integer chargingSize = deviceAlgorithmChargingHistory.getChargingSize();
            for (Integer integer = 0; integer < chargingSize; integer++) {
                //查询生效时间最大的算法计费
                LambdaQueryWrapper<DeviceAlgorithmCharging> lambdaQueryWrapper = new LambdaQueryWrapper<DeviceAlgorithmCharging>()
                        .eq(DeviceAlgorithmCharging::getDeviceId, deviceAlgorithmChargingHistory.getDeviceId())
                        .eq(DeviceAlgorithmCharging::getAlgorithmId, deviceAlgorithmChargingHistory.getAlgorithmId())
                        .eq(DeviceAlgorithmCharging::getType, type)
                        .orderByDesc(DeviceAlgorithmCharging::getBeginTime)
                        .last("limit 1");
                DeviceAlgorithmCharging deviceAlgorithmCharging = getOne(lambdaQueryWrapper);
                int beginTime;
                int timeout;
                if (deviceAlgorithmCharging == null) {
                    beginTime = Integer.valueOf(DataTime.toString(deviceAlgorithmChargingHistory.getCreateTime(), "yyyyMMdd"));
                    String stringAround = DataTime.getStringAround(0, 1, 0, 0, 0, 0, DataTime.toString(deviceAlgorithmChargingHistory.getCreateTime()));
                    timeout = Integer.valueOf(DataTime.toString(DataTime.toLocal(stringAround), "yyyyMMdd"));
                } else {
                    LocalDateTime localDateTime = DataTime.toLocal(deviceAlgorithmCharging.getBeginTime() + "01000000", "yyyyMMddHHmmss");
                    String s = DataTime.toString(localDateTime, "yyyy-MM-dd HH:mm:ss");
                    String stringAround = DataTime.getStringAround(0, 1, 0, 0, 0, 0, s);
                    beginTime = Integer.valueOf(DataTime.toString(DataTime.toLocal(stringAround), "yyyyMMdd"));
                    String stringAround1 = DataTime.getStringAround(0, 1, 0, 0, 0, 0, stringAround);
                    timeout = Integer.valueOf(DataTime.toString(DataTime.toLocal(stringAround1), "yyyyMMdd"));
                }
                DeviceAlgorithmCharging deviceAlgorithmChargingInfo = new DeviceAlgorithmCharging()
                        .setDeviceId(deviceAlgorithmChargingHistory.getDeviceId())
                        .setAlgorithmId(deviceAlgorithmChargingHistory.getAlgorithmId())
                        .setUnusedSize(moon.getSize())
                        .setType(type)
                        .setName(deviceAlgorithmChargingHistory.getName())
                        .setTimeout(timeout)
                        .setBeginTime(beginTime)
                        .setCreateTime(now)
                        .setUpdateTime(now);
                save(deviceAlgorithmChargingInfo);
            }
        });
        //处理无期限
        infs.forEach(deviceAlgorithmChargingHistory -> {
            DeviceAlgorithmCharging deviceAlgorithmChargingInfo = new DeviceAlgorithmCharging()
                    .setDeviceId(deviceAlgorithmChargingHistory.getDeviceId())
                    .setAlgorithmId(deviceAlgorithmChargingHistory.getAlgorithmId())
                    .setUnusedSize(deviceAlgorithmChargingHistory.getChargingSize())
                    .setType(deviceAlgorithmChargingHistory.getType())
                    .setName(deviceAlgorithmChargingHistory.getName())
                    .setTimeout(-1)
                    .setBeginTime(yyyyMMdd)
                    .setCreateTime(now)
                    .setUpdateTime(now);
            save(deviceAlgorithmChargingInfo);
        });
        //处理赠送
        gives.forEach(deviceAlgorithmChargingHistory -> {
            DeviceAlgorithmCharging deviceAlgorithmChargingInfo = new DeviceAlgorithmCharging()
                    .setDeviceId(deviceAlgorithmChargingHistory.getDeviceId())
                    .setAlgorithmId(deviceAlgorithmChargingHistory.getAlgorithmId())
                    .setUnusedSize(deviceAlgorithmChargingHistory.getChargingSize())
                    .setType(deviceAlgorithmChargingHistory.getType())
                    .setName(deviceAlgorithmChargingHistory.getName())
                    .setTimeout(payNotice.getTimeout())
                    .setBeginTime(yyyyMMdd)
                    .setCreateTime(now)
                    .setUpdateTime(now);
            save(deviceAlgorithmChargingInfo);
        });
        if (Emptys.check(deviceAlgorithmChargingHistories)) {
            Integer status = SysDictUtils.getValue(EnumDeviceAlgorithmChargingHistoryStatus.Code.CODE.getCode(), EnumDeviceAlgorithmChargingHistoryStatus.N_2.getCode(), Integer.class);
            deviceAlgorithmChargingHistories.forEach(deviceAlgorithmChargingHistory -> deviceAlgorithmChargingHistory.setStatus(status));
            deviceAlgorithmChargingHistoryService.updateBatchById(deviceAlgorithmChargingHistories);
        }
        return R.ok();
    }

    @Override
    @ApiOperation("扣费")
    public R<Map<Long, Integer>> charging(List<DeviceAlgorithmChargingDto.Charging> chargings) {
        //封装默认返回值
        Map<Long, Integer> map = new JHashMap<>(chargings.size());
        chargings.forEach(charging -> map.put(charging.getMercDeviceAlgorithmChargingId(), charging.getSize()));
        //查询数据
        List<DeviceAlgorithmCharging> queryList = new ArrayList<>(chargings.size());
        for (DeviceAlgorithmChargingDto.Charging charging : chargings) {
            DeviceAlgorithmCharging deviceAlgorithmCharging = new DeviceAlgorithmCharging()
                    .setDeviceId(charging.getDeviceId())
                    .setAlgorithmId(charging.getAlgorithmId())
                    .setBeginTime(charging.getAlgorithmDate());
            queryList.add(deviceAlgorithmCharging);
        }
        List<DeviceAlgorithmCharging> deviceAlgorithmChargingsList = baseMapper.chargingQuery(queryList);
        if (!Emptys.check(deviceAlgorithmChargingsList)) {
            return R.ok(map);
        }
        JList<DeviceAlgorithmCharging> deviceAlgorithmChargings = new JArrayList<>(deviceAlgorithmChargingsList);
        //查询计费类型字典
        Map<String, SysDictRedis> deviceAlgorithmChargingTypeMap = SysDictUtils.get(EnumDeviceAlgorithmChargingType.Code.CODE.getCode());
        SysDictRedis sysDictRedis1 = deviceAlgorithmChargingTypeMap.get(EnumDeviceAlgorithmChargingType.N_1.getCode());
        SysDictRedis sysDictRedis2 = deviceAlgorithmChargingTypeMap.get(EnumDeviceAlgorithmChargingType.N_2.getCode());
        SysDictRedis sysDictRedis3 = deviceAlgorithmChargingTypeMap.get(EnumDeviceAlgorithmChargingType.N_3.getCode());
        //循环扣费
        chargings.forEach(charging -> {
            //应扣数量
            int size = charging.getSize();
            //查找对应算法
            JList<DeviceAlgorithmCharging> deviceAlgorithmChargingJlist = deviceAlgorithmChargings.filter()
                    .eq(DeviceAlgorithmCharging::getDeviceId, charging.getDeviceId())
                    .eq(DeviceAlgorithmCharging::getAlgorithmId, charging.getAlgorithmId())
                    .list();
            if (!Emptys.check(deviceAlgorithmChargingJlist)) {
                return;
            }
            JMap<Integer, List<DeviceAlgorithmCharging>> deviceAlgorithmChargingJMap = deviceAlgorithmChargingJlist.toMap(DeviceAlgorithmCharging::getType).group();
            //递归扣费 扣费顺序为：赠送 -> 有期限 -> 无期限
            int size2 = charging(size, deviceAlgorithmChargingJMap, new JArrayList<SysDictRedis>().set(sysDictRedis3).set(sysDictRedis1).set(sysDictRedis2));
            map.put(charging.getMercDeviceAlgorithmChargingId(), size2);
        });
        return R.ok(map);
    }

    /**
     * 递归扣费
     *
     * @return 未扣条数
     */
    private int charging(int size, JMap<Integer, List<DeviceAlgorithmCharging>> deviceAlgorithmChargingJMap, List<SysDictRedis> sysDictRediss) {
        if (size > 0 && Emptys.check(sysDictRediss)) {
            SysDictRedis sysDictRedis = sysDictRediss.remove(0);
            LocalDateTime now = LocalDateTime.now();
            List<DeviceAlgorithmCharging> deviceAlgorithmChargings = deviceAlgorithmChargingJMap.get(Integer.valueOf(sysDictRedis.getValue()));
            if (Emptys.check(deviceAlgorithmChargings)) {
                for (DeviceAlgorithmCharging deviceAlgorithmCharging : deviceAlgorithmChargings) {
                    int unusedSize = deviceAlgorithmCharging.getUnusedSize();
                    if (unusedSize > 0) {
                        int newUnusedSize = unusedSize - size;
                        int makeSize;
                        //扣除后=负数则算法条数不足
                        if (newUnusedSize < 0) {
                            size = (~(newUnusedSize - 1));
                            newUnusedSize = 0;
                            makeSize = unusedSize;
                        } else {
                            makeSize = size;
                            size = 0;
                        }
                        DeviceAlgorithmCharging updateDeviceAlgorithmCharging = new DeviceAlgorithmCharging()
                                .setId(deviceAlgorithmCharging.getId())
                                .setUnusedSize(newUnusedSize)
                                .setMakeSize(deviceAlgorithmCharging.getMakeSize() + makeSize)
                                .setUpdateTime(now);
                        updateById(updateDeviceAlgorithmCharging);
                        deviceAlgorithmCharging.setUnusedSize(updateDeviceAlgorithmCharging.getUnusedSize())
                                .setMakeSize(updateDeviceAlgorithmCharging.getMakeSize());
                    }
                }
            }
            if (size > 0 && Emptys.check(sysDictRediss)) {
                return charging(size, deviceAlgorithmChargingJMap, sysDictRediss);
            }
        }
        return size;
    }
}